;	This program is free software; you can redistribute it and/or modify
;	it under the terms of the GNU General Public License as published by
;	the Free Software Foundation; either version 2 of the License, or
;	(at your option) any later version.
;
;	This program is distributed in the hope that it will be useful,
;	but WITHOUT ANY WARRANTY; without even the implied warranty of
;	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;	GNU General Public License for more details.
;
;	You should have received a copy of the GNU General Public License
;	along with this program; if not, write to the Free Software
;	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
Exponent:
	;Input - HL points to the base, DE points to the exponent
	;Output - HL points to the result of (HL)^(DE)
	inc hl
	ld a,(hl)
	and 80h
	dec hl
	ld b,a
	inc de
	ld a,(de)
	dec de
	bit 7,a
	jp nz,DomainError
	and b
	ld (tempN13),a			;sign of answer
	ld (tempN12),hl
	ld bc,(areaFreeStart)
	ld (tempN8),bc		;temp8 is the place to overwrite answers to at the end
	push hl
	ex de,hl
	call BCD_Clone
	ld (tempN9),hl		;temp9 is the exponent....
	call BCD_CheckZero
	push hl
	jp z,Exponent_WasZero
	pop hl
	pop hl
	call BCD_Clone
	;HL now is our b...
	ld (tempN11),hl		;temp11 is 'b'...the base...our exponent is in 
	call ReturnOne		;we need a 1 to start with
	ld (tempN10),hl		;temp10 is 'a'...the coefficient
Exponent_Loop:
	ld hl,(tempN9)
	call BCD_CheckZero		;is the exponent (c) zero?
	jp z,Exponent_AlmostDone
	ld c,(hl)
	inc hl
	ld b,(hl)
	add hl,bc
	;HL is the last byte...check it.
	ld a,(hl)
	and 1
	jr z,Exponent_Loop_Even		;exponent is even
	;The exponent is odd
	;a is equal to a * b
	ld hl,(tempN10)
	ld de,(tempN11)
	call BCD_Multiply
	bit flagModular,(iy+asm_Flag1)
	jr z,Exponent_NoMod1
	;Take the answer mod (tempN0)
	ld de,(tempN0)
	push hl
	call BCD_Division
	;DE is the answer...
	pop hl
	ex de,hl
	;DE is where to shift to
	;HL is the new answer
	ld c,(hl)
	inc hl
	ld b,(hl)
	res 7,b
	dec hl
	inc bc
	inc bc
	push de
	ldir
	ld (areaFreeStart),de
	pop hl
Exponent_NoMod1:
	;HL is the result...now we want to make this A...so save it, find the old A...
	;Remove the old A by shifting everything down over it....update all pointers accordingly
	ex de,hl
	;DE is result
	ld hl,(tempN10)
	ld c,(hl)
	inc hl
	ld b,(hl)
	res 7,b
	inc bc
	inc bc
	dec hl
	push hl	;HL was the old A
	ld hl,(areaFreeStart)
	or a
	sbc hl,bc
	ld (areaFreeStart),hl
	ex de,hl		;HL becomes the result...
	ld c,(hl)
	inc hl
	ld b,(hl)
	res 7,b
	inc bc
	inc bc
	dec hl
	pop de		;DE is the old A
	ldir			;shift down, and now we are done with that...
	ld hl,(tempN9)
	call BCD_Decrement		;decrease exponent
	jr Exponent_Loop			;loop back up
Exponent_Loop_Even:
	;We had an even exponent
	;We must set b = b^2
	ld hl,(tempN11)
	ld d,h
	ld e,l
	call BCD_Multiply		;b^2
	bit flagModular,(iy+asm_Flag1)
	jr z,Exponent_NoMod2
	ld de,(tempN0)
	call BCD_Division
	ex de,hl
Exponent_NoMod2:
	;HL is the result....we must now do something tricky...
	;Move A past the new B.
	;Then shift the B,A down until the B is in the same address as before.
	;Update all pointers!
	push hl		;HL is the saved result pointer...need it later for shifting down
	ld hl,(tempN10)
	ld c,(hl)
	inc hl
	ld b,(hl)
	res 7,b
	dec hl
	inc bc
	inc bc		;BC is total length...HL is start
	ld de,(areaFreeStart)		;where to copy too...but first add on
	push hl		;HL is 'a' source
	push de
	ld h,d
	ld l,e
	add hl,bc
	ld de,(areaFreeEnd)
	or a
	sbc hl,de
	jp nc,MemoryError
	add hl,de
	ld (areaFreeStart),hl
	pop de
	pop hl
	ld (tempN10),de		;update pointer for 'a'
	ldir			;shifted A past B...we updated the pointers
	;Now we must shift the whole thing down...from the pushed result pointer to
	;the byte of (areaFreeStart)-1. Then update pointers for 'a', and areaFreeStart.
	pop de		;DE is destination for it
	ld hl,(areaFreeStart)
	or a
	sbc hl,de		;difference to shift
	ld b,h
	ld c,l
	;DE still is the destination...
	ex de,hl
	ld de,(tempN11)
	;BC is the length of the total block...
	push hl
	ldir		;shift down
	;Now we must update pointers. Decrement 'a' pointer and (areaFreeStart) pointer by
	;the difference between the result pointer and the original pointer
	pop hl
	ld de,(tempN11)
	or a
	sbc hl,de
	ld b,h
	ld c,l
	ld hl,(tempN10)
	or a
	sbc hl,bc
	ld (tempN10),hl
	ld hl,(areaFreeStart)
	or a
	sbc hl,bc
	ld (areaFreeStart),hl	;all pointers updated now...we're done with this step!
	;We must divide the exponent by two now
	ld hl,(tempN9)
	call DivisionBy2
	jp Exponent_Loop		;loop back
Exponent_AlmostDone:
	bit flagModular,(iy+asm_Flag1)
	jr z,Exponent_NoMod3
	ld hl,(tempN10)
	ld de,(tempN0)
	call BCD_Division
	ld (tempN10),de
Exponent_NoMod3:
	ld hl,(tempN10)
	ld de,(tempN8)
	ld c,(hl)
	inc hl
	ld b,(hl)
	res 7,b
	dec hl
	inc bc
	inc bc
	push bc
	ldir		;copy
	pop bc
	ld hl,(tempN8)
	add hl,bc
	ld (areaFreeStart),hl
	ld hl,(tempN8)
	inc hl
	ld a,(tempN13)
	xor (hl)
	ld (hl),a
	dec hl
	ret
Exponent_WasZero:
	pop de
	pop hl
	jp ReturnOne
