;	This program is free software; you can redistribute it and/or modify
;	it under the terms of the GNU General Public License as published by
;	the Free Software Foundation; either version 2 of the License, or
;	(at your option) any later version.
;
;	This program is distributed in the hope that it will be useful,
;	but WITHOUT ANY WARRANTY; without even the implied warranty of
;	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;	GNU General Public License for more details.
;
;	You should have received a copy of the GNU General Public License
;	along with this program; if not, write to the Free Software
;	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

; Essential math routines
BCD_Negate:
	;Negate a number...HL points to it
	ld c,(hl)
	inc hl
	ld a,(hl)
	xor 80h
	ld (hl),a
	ld b,a
	res 7,b
BCD_Negate_Go:
	add hl,bc
	or a
	push af
BCD_Negate_Loop:
	pop af
	ld a,0
	sbc (hl)
	daa
	push af
	ld (hl),a
	dec hl
	dec bc
	ld a,b
	or c
	jr nz,BCD_Negate_Loop
	pop af
	ret
BCD_Decrement:
	;Decrements a number in HL by 1
	ld c,(hl)
	inc hl
	ld b,(hl)
	res 7,b
	add hl,bc
	scf
	push af
BCD_Decrement_Loop:
	pop af
	ld a,(hl)
	sbc a,0
	daa
	ld (hl),a
	push af
	dec hl
	dec bc
	ld a,b
	or c
	jr nz,BCD_Decrement_Loop
	pop af
	ret
BCD_CheckZero:
	;Input - HL, number to check for being zero
	;Output - HL, unchanged. Z if zero, NZ otherwise
	push hl
	ld c,(hl)
	inc hl
	ld b,(hl)
	res 7,b
	inc hl
BCD_CheckZero_Loop:
	ld a,(hl)
	or a
	jr nz,BCD_CheckZero_Done
	inc hl
	dec bc
	ld a,b
	or c
	jr nz,BCD_CheckZero_Loop
BCD_CheckZero_Done:
	pop hl
	ret
BCD_CheckOne:
	;Input - HL number to check for being one
	;Output - HL, unchanged. Z if one, NZ otherwise
	push hl
	ld c,(hl)
	inc hl
	ld b,(hl)
	ld hl,1
	or a
	sbc hl,bc
	jr nz,BCD_CheckOne_RetNZ
	pop hl
	inc hl
	inc hl
	ld a,(hl)
	dec hl
	dec hl
	dec a
	ret
BCD_CheckOne_RetNZ:
	pop hl
	ret
DivisionBy2:
	;Input - HL is a number, it's divided by 2
	;Output - HL points to same number
	push hl
	ld c,(hl)
	inc hl
	ld b,(hl)
	res 7,b
	inc hl
	ld de,TwoTable
DivisionBy2_Loop:
	push hl
	ld l,(hl)
	ld h,0
	add hl,de
	ld a,(hl)
	;A now is the divided by two value
	pop hl
	ld e,(hl)
	ld (hl),a
	bit 0,e
	jr nz,DivisionBy2_1
	ld de,TwoTable
	jr DivisionBy2_Next
DivisionBy2_1:
	ld de,TwoTable_Odd
DivisionBy2_Next:
	inc hl
	dec bc
	ld a,b
	or c
	jr nz,DivisionBy2_Loop
	pop hl
	ret
TwoTable:
	DB 0h, 0h, 1h, 1h, 2h, 2h, 3h, 3h, 4h, 4h, 0h, 0h, 0h, 0h, 0h, 0h, 5h
	DB 5h, 6h, 6h, 7h, 7h, 8h, 8h, 9h, 9h, 0h, 0h, 0h, 0h, 0h, 0h, 10h, 10h
	DB 11h, 11h, 12h, 12h, 13h, 13h, 14h, 14h, 0h, 0h, 0h, 0h, 0h, 0h, 15h, 15h, 16h
	DB 16h, 17h, 17h, 18h, 18h, 19h, 19h, 0h, 0h, 0h, 0h, 0h, 0h, 20h, 20h, 21h, 21h
	DB 22h, 22h, 23h, 23h, 24h, 24h, 0h, 0h, 0h, 0h, 0h, 0h, 25h, 25h, 26h, 26h, 27h
	DB 27h, 28h, 28h, 29h, 29h, 0h, 0h, 0h, 0h, 0h, 0h, 30h, 30h, 31h, 31h, 32h, 32h
	DB 33h, 33h, 34h, 34h, 0h, 0h, 0h, 0h, 0h, 0h, 35h, 35h, 36h, 36h, 37h, 37h, 38h
	DB 38h, 39h, 39h,53h,74h,65h,70h,68h,61h, 40h, 40h, 41h, 41h, 42h, 42h, 43h, 43h
	DB 44h, 44h,6Eh,69h,65h, 0h, 0h, 0h, 45h, 45h, 46h, 46h, 47h, 47h, 48h, 48h, 49h,49h
TwoTable_Odd:
	DB 50h, 50h, 51h, 51h, 52h, 52h, 53h, 53h, 54h, 54h, 0h, 0h, 0h, 0h, 0h, 0h, 55h
	DB 55h, 56h, 56h, 57h, 57h, 58h, 58h, 59h, 59h, 0h, 0h, 0h, 0h, 0h, 0h, 60h, 60h
	DB 61h, 61h, 62h, 62h, 63h, 63h, 64h, 64h, 0h, 0h, 0h, 0h, 0h, 0h, 65h, 65h, 66h
	DB 66h, 67h, 67h, 68h, 68h, 69h, 69h, 0h, 0h, 0h, 0h, 0h, 0h, 70h, 70h, 71h, 71h
	DB 72h, 72h, 73h, 73h, 74h, 74h, 0h, 0h, 0h, 0h, 0h, 0h, 75h, 75h, 76h, 76h, 77h
	DB 77h, 78h, 78h, 79h, 79h, 0h, 0h, 0h, 0h, 0h, 0h, 80h, 80h, 81h, 81h, 82h, 82h
	DB 83h, 83h, 84h, 84h, 0h, 0h, 0h, 0h, 0h, 0h, 85h, 85h, 86h, 86h, 87h, 87h, 88h
	DB 88h, 89h, 89h,56h,61h,6Eh,65h,73h,73h, 90h, 90h, 91h, 91h, 92h, 92h, 93h, 93h
	DB 94h, 94h,61h, 0h, 0h, 0h, 0h, 0h, 95h, 95h, 96h, 96h, 97h, 97h, 98h, 98h, 99h,99h
ReturnOne:
	;Inputs - none...
	;Output - makes the number one, and returns HL as pointer :)
	ld a,1
	jr ReturnX
ReturnZero:
	;Inputs - none...
	;Output - makes the number 0
	xor a
ReturnX:
	;Inputs - A is a single digit
	;Output - makes that number
	push af
	ld de,(areaFreeStart)
	ld hl,3
	add hl,de
	ld bc,(areaFreeEnd)
	or a
	sbc hl,bc
	jp nc,MemoryError
	add hl,bc
	ld (areaFreeStart),hl
	ld hl,ReturnZero_Data
	pop af
	push de
	ld bc,3
	ldir
	dec de
	ld (de),a
	pop hl
	ret
ReturnOne_Data:
	DW 0001h
	DB 01h
ReturnZero_Data:
	DW 0001h
	DB 00h
BCD_Clone:
	;Input - HL points to number to clone
	;Output - HL points to output
	ld c,(hl)
	inc hl
	ld b,(hl)
	res 7,b
	inc bc
	inc bc
	push bc
	push hl
	ld hl,(areaFreeStart)
	add hl,bc
	ld bc,(areaFreeEnd)
	or a
	sbc hl,bc
	jp nc,MemoryError
	pop hl
	pop bc
	dec hl
	ld de,(areaFreeStart)
	push de
	ldir		;copy
	ld (areaFreeStart),de
	pop hl
	ret
BCD_Compare:
	;Compares two numbers to see which is larger
	;Input - HL, DE
	;Outputs:
	;	If HL > DE, then NC, NZ
	;	If HL < DE, then C, NZ
	;	If HL = DE, then NC, Z
	; HL, DE preserved
	push hl
	push de
	push hl
	ld c,(hl)
	inc hl
	ld b,(hl)
	res 7,b		;BC is length of (HL)
	ld a,(de)
	ld l,a
	inc de
	ld a,(de)
	res 7,a
	ld h,a	;HL is length of (DE)
	;Check the lengths first
	or a
	sbc hl,bc
	jr c,BCD_Compare_HLGreaterThanDE
	jr z,BCD_Compare_EqualLengths
	;HL<DE
	or 1
BCD_Compare_HLGreaterThanDE:
	ccf
	pop hl
	pop de
	pop hl
	ret
BCD_Compare_EqualLengths:
	;If they are the same lengths, compare byte by byte
	pop hl
	ld c,(hl)
	inc hl
	ld b,(hl)
	res 7,b
	inc hl
	inc de
	ex de,hl
BCD_Compare_Loop:
	ld a,(de)
	cp (hl)
	jr c,BCD_Compare_Over
	jr nz,BCD_Compare_Over
	;they match...go again
	inc hl
	inc de
	dec bc
	ld a,b
	or c
	jr nz,BCD_Compare_Loop
	;This will be zero if it's gone through all the bytes and they match...works perfectly
BCD_Compare_Over:
	pop de
	pop hl
	ret
GCD:
	;Greatest common divisor of two numbers
	;Inputs - HL, DE
	;Outputs - HL
	;HL = a, DE = b
	;algorithm:
	;if b=0, then done
	;otherwise { c = a mod b, a = b, b = c }
	ld (temp14),hl		;location of A
	ld (temp15),de		;location of B
GCD_Loop:
	ld hl,(temp15)
	call BCD_CheckZero
	jr z,GCD_Done
	;b=0 complete.
	;now c = a mod b
	;after a division, the divisor is fine
	ld hl,(temp14)
	call BCD_Clone
	ld de,(temp15)
	call BCD_Division
	;a/b...DE is the remainder
	;we have [a][b][a'][c quotient/c remainder]
	push de			;temp16 is c
	ld hl,(temp15)		;get b...must set a equal to b
	call BCD_Clone		;HL is a copy of b...
	ld (temp14),hl		;which becomes the new a....
	;now b = c....
	pop hl
	ld (temp15),hl		;b = c
	jr GCD_Loop
GCD_Done:
	ld hl,(temp14)
	ret
LCM:
	;Calculates the least common multiple of two numbers
	;Inputs - HL, DE
	;Outputs - lcm in HL
	;algorithm:
	;gcd(a,b) * lcm(a,b) = a * b
	push de
	push hl
	push de
	call BCD_Clone
	ld (temp16),hl
	call BCD_CheckZero
	jp z,LCM_ThreeStack
	pop hl
	call BCD_Clone
	ld (temp17),hl		;temp16, temp17 are the original values
	call BCD_CheckZero
	jp z,LCM_TwoStack
	pop hl
	pop de
	call BCD_Multiply
	push hl
	ld hl,(temp16)
	ld de,(temp17)
	call GCD
	ex de,hl
	pop hl
	jp BCD_Division
LCM_ThreeStack:
	pop hl
LCM_TwoStack:
	pop hl
	pop hl
	jp ReturnZero
ClearZeroes:
	;Input - HL points to a number
	;Output - HL points to the number, minus any superfluous prepended zeroes
	push hl
	ld c,(hl)
	inc hl
	ld b,(hl)
	res 7,b
	inc hl
ClearZeroes_Loop:
	ld a,(hl)
	or a
	jr nz,ClearZeroes_NonZero
	inc hl
	dec bc
	ld a,b
	or c
	jr nz,ClearZeroes_Loop
	pop hl
	ld (hl),1
	inc hl
	ld (hl),0
	inc hl
	ld (hl),0
	ret
ClearZeroes_NonZero:
	dec hl
	ld (hl),b
	dec hl
	ld (hl),c
	pop bc
	ret
TenX:
	;Input - HL points to X
	;Output - returns 10^X
	inc hl
	ld a,(hl)
	or a
	jp nz,MemoryError
	bit 7,a
	jp nz,DomainError
	dec hl
	ld de,TenX_65535
	call BCD_Compare
	jp nc,MemoryError
	ld c,(hl)			;C is the length in bytes
	ld b,0
	inc hl
	add hl,bc
	ld ix,0
	ld de,1
	ld (temp1),de
	ld de,10
	ld (temp2),de
	call TenX_Add
	dec hl
	dec c
	jr z,TenX_Ready
	ld de,100
	ld (temp1),de
	ld de,1000
	ld (temp2),de
	call TenX_Add
	dec hl
	dec c
	jr z,TenX_Ready
	ld de,10000
	ld (temp1),de
	call TenX_Add
TenX_Ready:
	push ix
	pop hl
	ld a,h
	or l
	jp z,ReturnOne
	inc hl
	push hl
	ld de,(areaFreeStart)
	srl h
	rr l
	ld bc,0
	adc hl,bc	
	push hl
	add hl,de
	ld bc,(areaFreeEnd)
	or a
	sbc hl,bc
	jp nc,MemoryError
	add hl,bc
	ld (areaFreeStart),hl
	ex de,hl
	;HL is where to put this
	pop bc
	ld (hl),c
	inc hl
	ld (hl),b
	inc hl
	ld a,10h
	pop de
	push hl
	bit 0,e
	jr z,TenX_NoShift
	rrca
	rrca
	rrca
	rrca
TenX_NoShift:
	ld (hl),a
	inc hl
	dec bc
	ld a,b
	or c
	jr z,TenX_Exit
	ld e,0
TenX_FillZero:
	ld (hl),e
	inc hl
	dec bc
	ld a,b
	or c
	jr nz,TenX_FillZero
TenX_Exit:
	pop hl
	dec hl
	dec hl
	ret
TenX_65535:
	DW 3
	DB 06h,55h,35h
TenX_Add:
	ld a,(hl)
	and 00001111b
	ld b,a
	ld de,(temp1)
	call TenX_Add_2
	ld de,(temp2)
	ld b,(hl)
	srl b
	srl b
	srl b
	srl b
TenX_Add_2:
	ld a,b
	or a
	ret z
TenX_Add2:
	add ix,de
	djnz TenX_Add2
	ret

